from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_config.const.types import DictConfig
from wexample_filestate.testing.test_abstract_operation import TestAbstractOperation
from wexample_filestate_git.test.mixin.test_git_state_manager_mixin import (
    TestGitFileStateManagerMixin,
)

if TYPE_CHECKING:
    from wexample_config.const.types import DictConfig


class TestGitRemoteCreateOperation(TestGitFileStateManagerMixin, TestAbstractOperation):
    def _operation_get_count(self) -> int:
        return 3  # One operation for each remote, 1 operation to create local repo

    def _operation_test_assert_applied(self) -> None:
        # Verify that repositories were created using the new interface
        self.mock_github_create_if_not_exists.assert_called_once_with(
            "https://github.com/test-namespace/test-repo.git"
        )

        self.mock_gitlab_create_if_not_exists.assert_called_once_with(
            "https://gitlab.com/test-namespace/test-repo.git"
        )

    def _operation_test_assert_initial(self) -> None:
        # Verify initial state - no repositories should exist
        self.mock_github_create_if_not_exists.assert_not_called()
        self.mock_gitlab_create_if_not_exists.assert_not_called()

    def _operation_test_assert_rollback(self) -> None:
        # Add rollback assertions if needed
        pass

    def _operation_test_setup(self) -> None:
        from unittest.mock import patch

        from wexample_filestate_git.remote.github_remote import GithubRemote
        from wexample_filestate_git.remote.gitlab_remote import GitlabRemote

        # Setup all mocks
        self.mock_github_connect = patch.object(GithubRemote, "connect").start()
        self.mock_github_create_if_not_exists = patch.object(
            GithubRemote, "create_repository_if_not_exists"
        ).start()
        self.mock_gitlab_connect = patch.object(GitlabRemote, "connect").start()
        self.mock_gitlab_create_if_not_exists = patch.object(
            GitlabRemote, "create_repository_if_not_exists"
        ).start()

        # Configure default mock behaviors
        self.mock_github_create_if_not_exists.return_value = False
        self.mock_gitlab_create_if_not_exists.return_value = False

        # Call parent setup after mocks are ready
        super()._operation_test_setup()

    def _operation_test_setup_configuration(self) -> DictConfig | None:
        from wexample_filestate.const.disk import DiskItemType

        self._remove_test_git_dir()
        self._mock_git_env()

        return {
            "children": [
                {
                    "name": "test_git_dir",
                    "should_exist": True,
                    "type": DiskItemType.DIRECTORY,
                    "git": {
                        "remote": [
                            {
                                "name": "github",
                                "type": "github",
                                "url": "https://github.com/test-namespace/test-repo.git",
                                "create_remote": True,
                            },
                            {
                                "name": "gitlab",
                                "type": "gitlab",
                                "url": "https://gitlab.com/test-namespace/test-repo.git",
                                "create_remote": True,
                            },
                        ]
                    },
                }
            ]
        }



from __future__ import annotations

from types import UnionType
from typing import TYPE_CHECKING

from wexample_config.config_option.abstract_list_config_option import (
    AbstractListConfigOption,
)

if TYPE_CHECKING:
    from types import UnionType


class RemoteConfigOption(AbstractListConfigOption):
    def get_item_class_type(self) -> type | UnionType:
        from wexample_filestate_git.config_option.remote_item_config_option import (
            RemoteItemConfigOption,
        )

        return RemoteItemConfigOption



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_filestate.operation.mixin.file_manipulation_operation_mixin import (
    FileManipulationOperationMixin,
)
from wexample_filestate_git.operation.abstract_git_operation import AbstractGitOperation

if TYPE_CHECKING:
    from wexample_config.config_option.abstract_config_option import (
        AbstractConfigOption,
    )
    from wexample_filestate.operation.abstract_operation import AbstractOperation


class GitInitOperation(FileManipulationOperationMixin, AbstractGitOperation):
    _has_initialized_git: bool = False

    def applicable_for_option(self, option: AbstractConfigOption) -> bool:
        from wexample_filestate_git.config_option.git_config_option import (
            GitConfigOption,
        )
        from wexample_helpers_git.helpers.git import git_is_init

        if not self._is_active_git_option(option):
            return False

        assert isinstance(option, GitConfigOption)
        return option.should_have_git() and not git_is_init(self.target.get_path())

    def apply(self) -> None:
        from git import Repo

        path = self.target.get_path()
        self._has_initialized_git = True

        repo = Repo.init(path)
        repo.init()

    def dependencies(self) -> list[type[AbstractOperation]]:
        from wexample_filestate.operation.file_create_operation import (
            FileCreateOperation,
        )

        return [FileCreateOperation]

    def describe_after(self) -> str:
        return "Initialized .git directory"

    def describe_before(self) -> str:
        return "No initialized .git directory"

    def description(self) -> str:
        return "Initialize .git directory"

    def undo(self) -> None:
        import shutil

        from wexample_helpers.const.globals import DIR_GIT

        if self._has_initialized_git:
            shutil.rmtree(self.target.get_path() / DIR_GIT)



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_filestate.operation.mixin.file_manipulation_operation_mixin import (
    FileManipulationOperationMixin,
)
from wexample_filestate_git.operation.abstract_git_operation import AbstractGitOperation

if TYPE_CHECKING:
    from wexample_config.config_option.abstract_config_option import (
        AbstractConfigOption,
    )


class GitCreateBranchOperation(FileManipulationOperationMixin, AbstractGitOperation):
    def applicable_for_option(self, option: AbstractConfigOption) -> bool:
        # Only applicable when git is enabled and repo exists
        if not self._is_active_git_option(option):
            return False

        if not option.should_have_git():
            return False

        # Determine desired branch name from config
        branch = self._get_desired_branch_name()
        if not branch:
            return False

        repo = self._get_target_git_repo()

        # Apply only if the branch does not exist locally
        return not any(h.name == branch for h in getattr(repo, "heads", []))

    def apply(self) -> None:
        """Create the desired branch locally if it does not exist.

        Notes:
        - This operation is local-only (no remote push, no deletion/rename of other branches).
        - The branch is created at current HEAD.
        """
        branch = self._get_desired_branch_name()
        if not branch:
            return

        repo = self._get_target_git_repo()
        if any(h.name == branch for h in getattr(repo, "heads", [])):
            return  # Already exists

        # Create new branch at current HEAD
        repo.create_head(branch)

    def dependencies(self):
        from wexample_filestate_git.operation.git_init_operation import GitInitOperation

        # Ensure repository is initialized before creating a branch
        return [GitInitOperation]  # type: ignore[return-value]

    def describe_after(self) -> str:
        branch = self._get_desired_branch_name() or "<unknown>"
        return f"Local branch created: {branch}"

    def describe_before(self) -> str:
        branch = self._get_desired_branch_name() or "<unknown>"
        return f"Local branch missing: {branch}"

    def description(self) -> str:
        return "Create local Git branch if missing"

    def undo(self) -> None:
        # No destructive undo: we do not auto-delete newly created branches
        pass

    # --- helpers ---
    def _get_desired_branch_name(self) -> str | None:
        """Resolve desired branch name from config.

        Accepts either a list (first item) or a single string value.
        Defaults to "main" if the option exists but is empty/invalid.
        """
        from wexample_filestate_git.config_option.git_config_option import (
            GitConfigOption,
        )
        from wexample_filestate_git.config_option.main_branch_config_option import (
            MainBranchConfigOption,
        )

        git_option = self.target.get_option(GitConfigOption)
        if not git_option:
            return None

        main_opt = git_option.get_option(MainBranchConfigOption)
        if not main_opt:
            return None

        v = main_opt.get_value()

        # Prefer list form
        if v.is_list():
            items = v.get_list()
            first = items[0]
            if isinstance(first, str):
                return first

            # If not plain string, try to coerce via build_value
            return self._build_str_value(first)

        # Fallback: single string value
        if v.is_str():
            return v.get_str()

        if isinstance(v, str):
            return v

        # If the option exists but has no usable value, default to "main"
        return "main"



from __future__ import annotations

import re

import requests
from wexample_helpers.classes.field import public_field
from wexample_helpers.decorator.base_class import base_class

from .abstract_remote import AbstractRemote


@base_class
class GitlabRemote(AbstractRemote):
    api_token: str = public_field(description="GitHub API token")
    base_url: str = public_field(
        default="https://gitlab.com/api/v4", description="GitLab API base URL"
    )

    def __attrs_post_init__(self) -> None:
        self.default_headers.update({"PRIVATE-TOKEN": self.api_token})

    @classmethod
    def build_remote_api_url_from_repo(cls, remote_url: str) -> str | None:
        """Build API base URL from a GitLab remote URL.

        Supports both:
        - https://gitlab.example.com/owner/repo.git -> https://gitlab.example.com/api/v4
        - ssh://git@gitlab.example.com:4567/owner/repo.git -> https://gitlab.example.com/api/v4
        - git@gitlab.example.com:owner/repo.git -> https://gitlab.example.com/api/v4
        """
        host = None
        # ssh:// URL form
        m = re.search(r"^ssh://[^@]+@([^/:]+)", remote_url)
        if m:
            host = m.group(1)
        # git@host:path form
        if host is None:
            m = re.search(r"^git@([^:]+):", remote_url)
            if m:
                host = m.group(1)
        # https://host/ form
        if host is None:
            m = re.search(r"^https?://([^/]+)/", remote_url)
            if m:
                host = m.group(1)
        if not host:
            return None
        return f"https://{host}/api/v4"

    @classmethod
    def detect_remote_type(cls, remote_url: str) -> bool:
        # Support both gitlab.com and custom GitLab instances
        return bool(re.search(r"gitlab\.[a-zA-Z0-9.-]+[:/]", remote_url))

    def check_connection(self) -> bool:
        try:
            url = f"{self.base_url.rstrip('/')}/user"

            response = requests.head(
                url, headers=self.default_headers, timeout=self.timeout
            )

            return response.status_code == 200

        except requests.exceptions.RequestException as e:
            return False

    def check_repository_exists(self, name: str, namespace: str) -> bool:
        """
        Check if a repository exists in the specified namespace.

        Args:
            name: Repository name
            namespace: Organization or user name (mandatory)
        """
        endpoint = f"projects/{namespace}%2F{name}"
        response = self.make_request(
            endpoint=endpoint,
            call_origin=__file__,
            expected_status_codes=[200, 404],
        )
        return response.status_code == 200

    def create_repository(
        self, name: str, namespace: str, description: str = "", private: bool = False
    ) -> dict:
        """
        Create a new repository in the specified namespace.

        Args:
            name: Repository name
            namespace: Organization or user name (mandatory)
            description: Optional repository description
            private: Whether the repository should be private
        """
        from wexample_helpers_api.enums.http import HttpMethod

        data = {
            "name": name,
            "path": name,
            "description": description,
            "visibility": "private" if private else "public",
            "initialize_with_readme": False,
            "namespace_id": self._get_namespace_id(namespace),
        }

        response = self.make_request(
            method=HttpMethod.POST, endpoint="projects", data=data, call_origin=__file__
        )
        return response.json()

    def create_repository_if_not_exists(
        self, remote_url: str, description: str = "", private: bool = False
    ) -> dict:
        """
        Create a repository from a complete remote URL if it doesn't exist.

        Args:
            remote_url: Complete GitLab repository URL
            description: Optional repository description
            private: Whether the repository should be private
        """
        repo_info = self.parse_repository_url(remote_url)

        if not self.check_repository_exists(repo_info["name"], repo_info["namespace"]):
            return self.create_repository(
                name=repo_info["name"],
                namespace=repo_info["namespace"],
                description=description,
                private=private,
            )
        return {}

    def parse_repository_url(self, remote_url: str) -> dict[str, str]:
        """
        Parse a GitLab repository URL to extract repository information.
        Supports both HTTPS and SSH URLs and custom GitLab instances:
        - https://gitlab.example.com/owner/repo.git
        - git@gitlab.example.com:owner/repo.git
        """
        # Extract the path part after the domain
        if remote_url.startswith("git@"):
            path = remote_url.split(":", 1)[1]
        else:
            # For HTTPS URLs, split on the third slash to get the path
            parts = remote_url.split("/", 3)
            path = parts[3] if len(parts) > 3 else ""

        # Remove .git suffix if present
        path = path.replace(".git", "")

        # Split the path into parts and extract name and namespace
        url_parts = path.split("/")
        if len(url_parts) >= 2:
            repo_name = url_parts[-1]
            namespace = url_parts[-2]
            return {"name": repo_name, "namespace": namespace}

        return {"name": url_parts[0], "namespace": ""}

    def _get_namespace_id(self, namespace_path: str) -> int | None:
        try:
            response = requests.get(
                f"{self.base_url.rstrip('/')}/namespaces",
                params={"search": namespace_path},
                headers=self.default_headers,
                timeout=self.timeout,
            )
            if response.status_code == 200:
                namespaces = response.json()
                for ns in namespaces:
                    if ns["path"] == namespace_path:
                        return ns["id"]
            return None
        except requests.exceptions.RequestException:
            return None



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_config.options_provider.abstract_options_provider import (
    AbstractOptionsProvider,
)

if TYPE_CHECKING:
    from wexample_config.config_option.abstract_config_option import (
        AbstractConfigOption,
    )


class GitConfigOptionsProvider(AbstractOptionsProvider):
    @classmethod
    def get_options(cls) -> list[type[AbstractConfigOption]]:
        from wexample_filestate.option.active_option import (
            ActiveOption,
        )
        from wexample_filestate_git.config_option.main_branch_config_option import (
            MainBranchConfigOption,
        )
        from wexample_filestate_git.config_option.remote_config_option import (
            RemoteConfigOption,
        )

        return [
            ActiveOption,
            MainBranchConfigOption,
            RemoteConfigOption,
        ]



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from wexample_filestate.operation.mixin.file_manipulation_operation_mixin import (
    FileManipulationOperationMixin,
)
from wexample_filestate_git.operation.abstract_git_operation import AbstractGitOperation

if TYPE_CHECKING:
    from git import Repo
    from wexample_config.config_option.abstract_config_option import (
        AbstractConfigOption,
    )
    from wexample_filestate.operation.abstract_operation import AbstractOperation


class GitRemoteAddOperation(FileManipulationOperationMixin, AbstractGitOperation):
    _created_remote: dict[str, bool]

    def __init__(self, **data) -> None:
        super().__init__(**data)
        self._created_remote = {}

    def applicable_for_option(self, option: AbstractConfigOption) -> bool:
        from wexample_filestate_git.config_option.git_config_option import (
            GitConfigOption,
        )
        from wexample_helpers_git.helpers.git import git_is_init

        if not self._is_active_git_option(option):
            return False

        assert isinstance(option, GitConfigOption)
        if option.should_have_git() and git_is_init(self.target.get_path()):
            value = self.target.get_option_value(GitConfigOption)
            if not value or not value.has_key_in_dict("remote"):
                return False

            return self._is_remote_missing_or_mismatched()
        return False

    def apply(self) -> None:
        from wexample_filestate.option.active_option import (
            ActiveOption,
        )
        from wexample_filestate_git.config_option.git_config_option import (
            GitConfigOption,
        )
        from wexample_helpers_git.helpers.git import git_remote_create_once

        value = self.target.get_option_value(GitConfigOption)

        if value.is_dict():
            for remote in value.get_dict().get("remote"):
                # Respect per-remote active flag (defaults to inactive if missing)
                if not self._is_active_flag(remote.get(ActiveOption.get_name())):
                    continue
                repo = self._get_target_git_repo()

                remote_name = self._build_str_value(remote["name"])
                remote_url = self._build_str_value(remote["url"])

                self._created_remote[remote_name] = (
                    git_remote_create_once(repo, remote_name, remote_url) is not None
                )

    def dependencies(self) -> list[type[AbstractOperation]]:
        from wexample_filestate_git.operation.git_init_operation import GitInitOperation

        return [GitInitOperation]

    def describe_after(self) -> str:
        desc = self._remotes_description()
        if desc:
            return f"Remote added in .git directory: {desc}"
        return "Remote added in .git directory"

    def describe_before(self) -> str:
        desc = self._remotes_description()
        if desc:
            return f"Remote missing in .git directory: {desc}"
        return "Remote missing in .git directory"

    def description(self) -> str:
        return "Add remote in .git directory"

    def undo(self) -> None:
        from wexample_filestate_git.config_option.git_config_option import (
            GitConfigOption,
        )

        option = cast(GitConfigOption, self.target.get_option(GitConfigOption))

        config = option.get_value().get_dict()
        for remote in config.get("remote"):
            if self._created_remote:
                repo = self._get_target_git_repo()
                remote_name = (
                    self._build_value(remote["name"])
                    if not isinstance(remote["name"], str)
                    else remote["name"]
                )

                if self._created_remote[remote_name] is True:
                    repo.delete_remote(remote=repo.remote(name=remote_name))

    def _get_target_git_repo(self) -> Repo:
        from git import Repo

        return Repo(self.target.get_path())

    def _is_remote_missing_or_mismatched(self) -> bool:
        from wexample_filestate.option.active_option import (
            ActiveOption,
        )
        from wexample_filestate_git.config_option.git_config_option import (
            GitConfigOption,
        )

        value = self.target.get_option_value(GitConfigOption)

        # If no value or no remotes key, we consider no need to apply here.
        if not value or not value.has_key_in_dict("remote"):
            return False

        # If any configured remote is missing by name or has a different URL, apply.
        repo = self._get_target_git_repo()

        configured_remotes = value.get_dict().get("remote", [])
        existing_by_name = {r.name: r for r in repo.remotes}

        for remote in configured_remotes:
            # Skip inactive remotes; treat missing flag as inactive
            if not self._is_active_flag(remote.get(ActiveOption.get_name())):
                continue
            desired_name = self._build_value(remote.get("name"))
            desired_url = self._build_value(remote.get("url"))

            if not desired_name:
                # Malformed config, skip this entry
                continue

            existing = existing_by_name.get(str(desired_name))
            if existing is None:
                return True

            if desired_url:
                existing_urls = {u for u in existing.urls}
                if str(desired_url) not in existing_urls:
                    return True

        # All configured remotes exist with expected URLs
        return False

    def _remotes_description(self) -> str:
        from wexample_filestate_git.config_option.git_config_option import (
            GitConfigOption,
        )

        value = self.target.get_option_value(GitConfigOption)

        if not value or not value.is_dict():
            return ""

        remotes = value.get_dict().get("remote")
        if not remotes:
            return ""

        parts: list[str] = []
        for remote in remotes:
            try:
                name = self._build_value(remote.get("name"))
                url = self._build_value(remote.get("url"))
                if name and url:
                    parts.append(f"{name} -> {url}")
                elif name:
                    parts.append(str(name))
                elif url:
                    parts.append(str(url))
            except Exception:
                # Be conservative: if anything goes wrong computing description, skip that entry
                continue

        return ", ".join(parts)



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_config.config_option.abstract_nested_config_option import (
    AbstractNestedConfigOption,
)

if TYPE_CHECKING:
    from wexample_config.options_provider.abstract_options_provider import (
        AbstractOptionsProvider,
    )


class RemoteItemConfigOption(AbstractNestedConfigOption):
    def get_options_providers(self) -> list[type[AbstractOptionsProvider]]:
        from wexample_filestate_git.options_provider.remote_item_options_provider import (
            RemoteItemOptionsProvider,
        )

        return [RemoteItemOptionsProvider]



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_config.options_provider.abstract_options_provider import (
    AbstractOptionsProvider,
)

if TYPE_CHECKING:
    from wexample_config.config_option.abstract_config_option import (
        AbstractConfigOption,
    )


class RemoteItemOptionsProvider(AbstractOptionsProvider):
    @classmethod
    def get_options(cls) -> list[type[AbstractConfigOption]]:
        from wexample_config.config_option.name_config_option import NameConfigOption
        from wexample_filestate.option.active_option import (
            ActiveOption,
        )
        from wexample_filestate.option.type_option import TypeOption
        from wexample_filestate_git.config_option.create_remote_config_option import (
            CreateRemoteConfigOption,
        )
        from wexample_filestate_git.config_option.url_config_option import (
            UrlConfigOption,
        )

        return [
            ActiveOption,
            NameConfigOption,
            UrlConfigOption,
            CreateRemoteConfigOption,
            TypeOption,
        ]



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_filestate.operation.abstract_operation import AbstractOperation

if TYPE_CHECKING:
    from git import Repo
    from wexample_config.config_option.abstract_config_option import (
        AbstractConfigOption,
    )
    from wexample_filestate.enum.scopes import Scope


class AbstractGitOperation(AbstractOperation):
    @classmethod
    def get_scope(cls) -> Scope:
        from wexample_filestate.enum.scopes import Scope

        return Scope.REMOTE

    # Shared Git helpers
    def _get_target_git_repo(self) -> Repo:
        """Return the GitPython Repo for the target path."""
        from git import Repo

        return Repo(self.target.get_path())

    # Evaluate an 'active' flag consistently across operations.
    # Accepts raw values (bool, int, str, etc.) and treats missing as inactive.
    def _is_active_flag(self, raw_value) -> bool:
        from wexample_filestate.option.active_option import (
            ActiveOption,
        )

        if raw_value is None:
            return True
        return ActiveOption.is_active(raw_value)

    def _is_active_git_option(self, option: AbstractConfigOption) -> bool:
        from wexample_filestate.option.active_option import (
            ActiveOption,
        )
        from wexample_filestate_git.config_option.git_config_option import (
            GitConfigOption,
        )

        if isinstance(option, GitConfigOption):
            value = self.target.get_option_value(GitConfigOption)
            if value:
                if value.has_key_in_dict(ActiveOption.get_name()):
                    active_option = value.get_dict().get(ActiveOption.get_name())
                    return ActiveOption.is_active(active_option)
                return True
        return False



from __future__ import annotations

from wexample_filestate.testing.abstract_state_manager_test import (
    AbstractStateManagerTest,
)
from wexample_filestate_git.test.mixin.test_git_state_manager_mixin import (
    TestGitFileStateManagerMixin,
)


class TestGitFileStateManager(TestGitFileStateManagerMixin, AbstractStateManagerTest):
    def test_setup(self) -> None:
        from wexample_filestate_git.operations_provider.git_operations_provider import (
            GitOperationsProvider,
        )
        from wexample_filestate_git.options_provider.git_options_provider import (
            GitOptionsProvider,
        )

        assert GitOptionsProvider in self.state_manager.options_providers
        assert GitOperationsProvider in self.state_manager.operations_providers



from __future__ import annotations

from typing import Any, Union

from wexample_config.config_option.abstract_config_option import AbstractConfigOption


class UrlConfigOption(AbstractConfigOption):
    @staticmethod
    def get_raw_value_allowed_type() -> Any:
        from collections.abc import Callable

        from wexample_filestate.item.abstract_item_target import AbstractItemTarget

        return Union[str, Callable[[AbstractItemTarget], str]]



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from wexample_config.options_provider.abstract_options_provider import (
    AbstractOptionsProvider,
)
from wexample_filestate.operations_provider.abstract_operations_provider import (
    AbstractOperationsProvider,
)

if TYPE_CHECKING:
    from pathlib import Path

    from wexample_config.options_provider.abstract_options_provider import (
        AbstractOptionsProvider,
    )
    from wexample_filestate.operations_provider.abstract_operations_provider import (
        AbstractOperationsProvider,
    )


class TestGitFileStateManagerMixin:
    _env_patcher = None

    def _get_git_dir_path(self, item_name: str) -> Path:
        """Get the path to the .git directory for a given item."""
        from wexample_helpers.const.globals import DIR_GIT

        return self.state_manager.find_by_name_or_fail(item_name).get_path() / DIR_GIT

    def _get_test_operations_providers(
        self,
    ) -> list[type[AbstractOperationsProvider]] | None:
        from wexample_filestate.operations_provider.default_operations_provider import (
            DefaultOperationsProvider,
        )
        from wexample_filestate_git.operations_provider.git_operations_provider import (
            GitOperationsProvider,
        )

        return [DefaultOperationsProvider, GitOperationsProvider]

    def _get_test_options_providers(
        self,
    ) -> list[type[AbstractOptionsProvider]] | None:
        from wexample_filestate.options_provider.default_options_provider import (
            DefaultOptionsProvider,
        )
        from wexample_filestate_git.options_provider.git_options_provider import (
            GitOptionsProvider,
        )

        return [DefaultOptionsProvider, GitOptionsProvider]

    def _mock_git_env(self) -> None:
        """Mock Git-related environment variables."""
        from unittest.mock import patch

        mock_env = {
            "GITHUB_API_TOKEN": "mock-token-123",
            "GITLAB_API_TOKEN": "mock-token-456",
            "GITHUB_DEFAULT_URL": "https://api.github.test",
            "GITLAB_DEFAULT_URL": "https://gitlab.test/api/v4",
        }

        self._env_patcher = patch.dict(os.environ, mock_env)
        self._env_patcher.start()

    def _remove_test_git_dir(self) -> None:
        from wexample_helpers.helpers.directory import directory_remove_tree_if_exists

        directory_remove_tree_if_exists(
            self._get_absolute_path_from_state_manager("test_git_dir/.git")
        )



from __future__ import annotations

from types import UnionType
from typing import TYPE_CHECKING

from wexample_config.config_option.abstract_list_config_option import (
    AbstractListConfigOption,
)

if TYPE_CHECKING:
    from types import UnionType


class MainBranchConfigOption(AbstractListConfigOption):
    @staticmethod
    def get_raw_value_allowed_type() -> type | UnionType:
        return str



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_config.options_provider.abstract_options_provider import (
    AbstractOptionsProvider,
)

if TYPE_CHECKING:
    from wexample_config.config_option.abstract_config_option import (
        AbstractConfigOption,
    )


class GitOptionsProvider(AbstractOptionsProvider):
    @classmethod
    def get_options(cls) -> list[type[AbstractConfigOption]]:
        from wexample_filestate_git.config_option.git_config_option import (
            GitConfigOption,
        )

        return [
            GitConfigOption,
        ]



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from wexample_filestate.operation.mixin.file_manipulation_operation_mixin import (
    FileManipulationOperationMixin,
)
from wexample_filestate_git.operation.abstract_git_operation import AbstractGitOperation

if TYPE_CHECKING:
    from wexample_config.config_option.abstract_config_option import (
        AbstractConfigOption,
    )
    from wexample_filestate.operation.abstract_operation import AbstractOperation
    from wexample_filestate_git.remote.abstract_remote import AbstractRemote


class GitRemoteCreateOperation(FileManipulationOperationMixin, AbstractGitOperation):
    @staticmethod
    def get_remote_types() -> list[type[AbstractRemote]]:
        from wexample_filestate_git.remote.github_remote import GithubRemote
        from wexample_filestate_git.remote.gitlab_remote import GitlabRemote

        return [GithubRemote, GitlabRemote]

    def applicable_for_option(self, option: AbstractConfigOption) -> bool:
        from wexample_filestate.option.active_option import (
            ActiveOption,
        )
        from wexample_filestate_git.config_option.create_remote_config_option import (
            CreateRemoteConfigOption,
        )
        from wexample_filestate_git.config_option.git_config_option import (
            GitConfigOption,
        )
        from wexample_filestate_git.config_option.remote_config_option import (
            RemoteConfigOption,
        )

        if not self._is_active_git_option(option):
            return False

        assert isinstance(option, GitConfigOption)

        git_option = self.target.get_option(GitConfigOption)
        remote_option = git_option.get_option(RemoteConfigOption)
        if not remote_option:
            return False

        # Trigger only if at least one remote needs to be created (doesn't exist yet)
        for remote_item_option in remote_option.children:
            # creation flag must be true
            create_remote_option = remote_item_option.get_option(
                CreateRemoteConfigOption
            )
            create_enabled = (
                create_remote_option is not None
                and create_remote_option.get_value().is_true()
            )

            # item must be active (missing flag => active by default)
            active_option = remote_item_option.get_option(ActiveOption)
            raw_active = (
                active_option.get_value().raw if active_option is not None else None
            )
            is_active = self._is_active_flag(raw_active)

            if create_enabled and is_active:
                # 3) resolve type and url
                resolved = self._resolve_remote_type_and_url(remote_item_option)
                if resolved:
                    remote_type, remote_url = resolved
                    # Derive API base URL from repo URL (supports custom domains)
                    remote_type.build_remote_api_url_from_repo(remote_url)
                    remote = self._build_remote_instance(
                        remote_type=remote_type, remote_url=remote_url
                    )
                    if remote:
                        remote.connect()
                        repo_info = remote.parse_repository_url(remote_url)
                        if not remote.check_repository_exists(
                            repo_info["name"], repo_info["namespace"]
                        ):
                            # At least one configured remote is missing: operation is applicable
                            return True

        return False

    def apply(self) -> None:
        from wexample_filestate_git.config_option.create_remote_config_option import (
            CreateRemoteConfigOption,
        )
        from wexample_filestate_git.config_option.git_config_option import (
            GitConfigOption,
        )
        from wexample_filestate_git.config_option.remote_config_option import (
            RemoteConfigOption,
        )
        from wexample_filestate_git.config_option.type_config_option import (
            TypeConfigOption,
        )
        from wexample_filestate_git.config_option.url_config_option import (
            UrlConfigOption,
        )
        from wexample_filestate_git.remote.github_remote import GithubRemote
        from wexample_filestate_git.remote.gitlab_remote import GitlabRemote
        from wexample_helpers_git.const.common import (
            GIT_PROVIDER_GITHUB,
            GIT_PROVIDER_GITLAB,
        )

        git_option = self.target.get_option(GitConfigOption)
        git_option.get_value()

        if git_option:
            git_option = cast(GitConfigOption, git_option)
            remote_option = cast(
                RemoteConfigOption, git_option.get_option(RemoteConfigOption)
            )

            if remote_option:
                for remote_item_option in remote_option.children:
                    remote_item_option = cast(
                        remote_option.get_item_class_type(), remote_item_option
                    )
                    create_remote_option = remote_item_option.get_option(
                        CreateRemoteConfigOption
                    )
                    url_option = remote_item_option.get_option(UrlConfigOption)
                    type_option = remote_item_option.get_option(TypeConfigOption)

                    if (
                        create_remote_option
                        and create_remote_option.get_value().is_true()
                    ):
                        # Support strings or callables in the UrlConfigOption value
                        remote_url = self._build_str_value(url_option.get_value())

                        # Auto-detect remote type if not specified
                        if type_option:
                            # Use specified type
                            type_map = {
                                GIT_PROVIDER_GITHUB: GithubRemote,
                                GIT_PROVIDER_GITLAB: GitlabRemote,
                            }
                            remote_type = type_map.get(
                                type_option.get_value().get_str().lower()
                            )
                        else:
                            # Auto-detect from URL
                            remote_type = self._detect_remote_type(remote_url)

                        if remote_type:
                            remote_type.build_remote_api_url_from_repo(remote_url)
                            remote = self._build_remote_instance(
                                remote_type=remote_type, remote_url=remote_url
                            )
                            if remote:
                                remote.connect()
                                # Create repository directly from URL
                                remote.create_repository_if_not_exists(remote_url)

    def dependencies(self) -> list[type[AbstractOperation]]:
        from wexample_filestate_git.operation.git_remote_add_operation import (
            GitRemoteAddOperation,
        )

        return [GitRemoteAddOperation]

    def describe_after(self) -> str:
        desc = self._create_remotes_description()
        if desc:
            return f"Remote repository created on remote platform: {desc}"
        return "Remote repository created on remote platform"

    def describe_before(self) -> str:
        desc = self._create_remotes_description()
        if desc:
            return f"Remote repository not created on remote platform: {desc}"
        return "Remote repository not created on remote platform"

    def description(self) -> str:
        return "Create remote repository on platform"

    def undo(self) -> None:
        # Note: We don't implement undo for remote repository creation
        # as it could be dangerous to automatically delete repositories
        pass

    def _build_remote_instance(
        self, remote_type: type[AbstractRemote], remote_url: str
    ) -> AbstractRemote | None:
        # Instantiate the proper remote with required constructor args
        # GithubRemote expects an api_token passed explicitly
        # We may find another way to pass tokens, with an option value.
        return remote_type(
            io=self.target.io,
            api_token=self.target.get_env_parameter(
                # GITHUB_API_TOKEN / GITLAB_API_TOKEN
                key=f"{remote_type.get_snake_short_class_name().upper()}_API_TOKEN"
            ),
            base_url=remote_type.build_remote_api_url_from_repo(remote_url),
        )

    def _create_remotes_description(self) -> str:
        from wexample_filestate_git.config_option.create_remote_config_option import (
            CreateRemoteConfigOption,
        )
        from wexample_filestate_git.config_option.git_config_option import (
            GitConfigOption,
        )
        from wexample_filestate_git.config_option.remote_config_option import (
            RemoteConfigOption,
        )
        from wexample_filestate_git.config_option.type_config_option import (
            TypeConfigOption,
        )
        from wexample_filestate_git.config_option.url_config_option import (
            UrlConfigOption,
        )
        from wexample_filestate_git.remote.github_remote import GithubRemote
        from wexample_filestate_git.remote.gitlab_remote import GitlabRemote

        parts: list[str] = []

        git_option = self.target.get_option(GitConfigOption)
        if not git_option:
            return ""
        remote_option = git_option.get_option(RemoteConfigOption)
        if not remote_option:
            return ""

        for remote_item_option in remote_option.children:
            create_remote_option = remote_item_option.get_option(
                CreateRemoteConfigOption
            )
            if not (
                create_remote_option and create_remote_option.get_value().is_true()
            ):
                continue

            url_option = remote_item_option.get_option(UrlConfigOption)
            type_option = remote_item_option.get_option(TypeConfigOption)

            remote_url = None
            remote_type_label = None

            if url_option:
                remote_url = self._build_str_value(url_option.get_value())

            if type_option:
                remote_type_label = type_option.get_value().get_str().lower()
            else:
                if remote_url:
                    detected = self._detect_remote_type(remote_url)
                    if detected is GithubRemote:
                        remote_type_label = "github"
                    elif detected is GitlabRemote:
                        remote_type_label = "gitlab"

            if remote_url and remote_type_label:
                parts.append(f"{remote_type_label}: {remote_url}")
            elif remote_url:
                parts.append(str(remote_url))
            elif remote_type_label:
                parts.append(str(remote_type_label))

        return ", ".join(parts)

    def _detect_remote_type(self, remote_url: str) -> type[AbstractRemote] | None:
        """
        Detect the remote type (GitHub, GitLab, etc.) from the URL.
        """
        for remote_type in self.get_remote_types():
            if remote_type.detect_remote_type(remote_url):
                return remote_type
        return None

    def _resolve_remote_type_and_url(self, remote_item_option):
        from wexample_filestate_git.config_option.type_config_option import (
            TypeConfigOption,
        )
        from wexample_filestate_git.config_option.url_config_option import (
            UrlConfigOption,
        )
        from wexample_filestate_git.remote.github_remote import GithubRemote
        from wexample_filestate_git.remote.gitlab_remote import GitlabRemote
        from wexample_helpers_git.const.common import (
            GIT_PROVIDER_GITHUB,
            GIT_PROVIDER_GITLAB,
        )

        url_option = remote_item_option.get_option(UrlConfigOption)
        type_option = remote_item_option.get_option(TypeConfigOption)

        if not url_option:
            return None

        remote_url = self._build_str_value(url_option.get_value())

        if type_option:
            type_map = {
                GIT_PROVIDER_GITHUB: GithubRemote,
                GIT_PROVIDER_GITLAB: GitlabRemote,
            }
            remote_type = type_map.get(type_option.get_value().get_str().lower())
        else:
            remote_type = self._detect_remote_type(remote_url)

        if not remote_type:
            return None

        return remote_type, remote_url



from __future__ import annotations

import re

from wexample_helpers.classes.field import public_field
from wexample_helpers.decorator.base_class import base_class

from .abstract_remote import AbstractRemote


@base_class
class GithubRemote(AbstractRemote):
    api_token: str = public_field(description="GitHub API token")
    base_url: str = public_field(
        default="https://api.github.com", description="GitHub API base URL"
    )

    def __attrs_post_init__(self) -> None:
        self.default_headers.update(
            {
                "Authorization": f"token {self.api_token}",
                "Accept": "application/vnd.github.v3+json",
            }
        )

    @classmethod
    def build_remote_api_url_from_repo(cls, remote_url: str) -> str | None:
        """Build API base URL from a GitHub remote URL.

        Supports:
        - https://github.com/owner/repo(.git)
        - git@github.com:owner/repo(.git)
        - GHES custom domains: https://github.example.com/... or git@github.example.com:...
        Returns the appropriate https://<host>/api/v3 for custom domains,
        or https://api.github.com for github.com.
        """
        m = re.search(r"^(?:https://|git@)([^/:]+)", remote_url)
        if not m:
            return None
        host = m.group(1)
        if host == "github.com":
            return "https://api.github.com"
        # GitHub Enterprise Server
        return f"https://{host}/api/v3"

    @classmethod
    def detect_remote_type(cls, remote_url: str) -> bool:
        return bool(re.search(r"github\.com[:/]", remote_url))

    def check_repository_exists(self, name: str, namespace: str) -> bool:
        """
        Check if a repository exists in the specified namespace.

        Args:
            name: Repository name
            namespace: Organization or user name (mandatory)
        """
        endpoint = f"repos/{namespace}/{name}"
        response = self.make_request(
            endpoint=endpoint,
            call_origin=__file__,
            expected_status_codes=[200, 404],
        )
        return response.status_code == 200

    def create_repository(
        self, name: str, namespace: str, description: str = "", private: bool = False
    ) -> dict:
        """
        Create a new repository in the specified namespace.

        Args:
            name: Repository name
            namespace: Organization or user name (mandatory)
            description: Optional repository description
            private: Whether the repository should be private
        """
        from wexample_helpers_api.enums.http import HttpMethod

        endpoint = f"orgs/{namespace}/repos"

        response = self.make_request(
            method=HttpMethod.POST,
            endpoint=endpoint,
            data={
                "name": name,
                "description": description,
                "private": private,
                "auto_init": True,
            },
            call_origin=__file__,
            expected_status_codes=[201],  # Only 201 Created is acceptable
            fatal_if_unexpected=True,  # Any other status code should raise an error
        )
        return response.json()

    def create_repository_if_not_exists(
        self, remote_url: str, description: str = "", private: bool = False
    ) -> dict:
        """
        Create a repository from a complete remote URL if it doesn't exist.

        Args:
            remote_url: Complete GitHub repository URL
            description: Optional repository description
            private: Whether the repository should be private
        """
        repo_info = self.parse_repository_url(remote_url)

        if not self.check_repository_exists(repo_info["name"], repo_info["namespace"]):
            return self.create_repository(
                name=repo_info["name"],
                namespace=repo_info["namespace"],
                description=description,
                private=private,
            )
        return {}

    def parse_repository_url(self, remote_url: str) -> dict[str, str]:
        """
        Parse a GitHub repository URL to extract repository information.
        Supports both HTTPS and SSH URLs:
        - https://github.com/owner/repo.git
        - git@github.com:owner/repo.git
        """
        # Remove protocol and domain
        path = re.sub(r"^(https://github\.com/|git@github\.com:)", "", remote_url)
        # Remove .git suffix if present
        path = path.replace(".git", "")

        parts = path.split("/")
        if len(parts) >= 2:
            return {"name": parts[-1], "namespace": parts[-2]}
        return {"name": parts[0], "namespace": ""}



from __future__ import annotations

import pytest
from wexample_filestate_git.remote.gitlab_remote import GitlabRemote
from wexample_filestate_git.testing.git_remote_test import GitRemoteTest


class TestGitlabRemote(GitRemoteTest):
    """Test cases for GitLab remote operations."""

    ENV_TOKEN_NAME = "GITLAB_API_TOKEN"
    SERVICE_CLASS = GitlabRemote

    @pytest.fixture
    def remote(self) -> GitlabRemote:
        from unittest.mock import patch

        from wexample_prompt.common.io_manager import IoManager

        io_manager = IoManager()
        with patch.dict("os.environ", {"GITLAB_API_TOKEN": "test_token"}):
            remote = GitlabRemote(io=io_manager)
            return remote

    def test_check_repository_exists(self, remote) -> None:
        from unittest.mock import patch

        with patch(
            "wexample_helpers_api.common.abstract_gateway.AbstractGateway.make_request"
        ) as mock_request:
            mock_request.return_value.status_code = 200

            exists = remote.check_repository_exists("test-repo", "test-namespace")

            mock_request.assert_called_once()
            self._assert_check_repository_exists_request(mock_request)
            assert exists is True

    def test_create_repository(self, remote) -> None:
        from unittest.mock import patch

        with patch(
            "wexample_helpers_api.common.abstract_gateway.AbstractGateway.make_request"
        ) as mock_request:
            mock_request.return_value.json.return_value = {"id": 1}

            result = remote.create_repository(
                name="test-repo",
                namespace="test-namespace",
                description="Test description",
                private=True,
            )

            mock_request.assert_called_once()
            self._assert_create_repository_request(mock_request)
            assert result == {"id": 1}

    def test_parse_repository_url_https(self, remote) -> None:
        url = "https://gitlab.com/test-namespace/test-repo.git"
        info = remote.parse_repository_url(url)
        assert info == {"name": "test-repo", "namespace": "test-namespace"}

    def test_parse_repository_url_ssh(self, remote) -> None:
        url = "git@gitlab.com:test-namespace/test-repo.git"
        info = remote.parse_repository_url(url)
        assert info == {"name": "test-repo", "namespace": "test-namespace"}

    def _assert_check_repository_exists_request(self, mock_request) -> None:
        """Assert the request parameters for checking repository existence."""
        assert (
            mock_request.call_args[1]["endpoint"]
            == f"projects/{self.test_namespace}%2F{self.test_repo_name}"
        )

    def _assert_create_repository_request(self, mock_request) -> None:
        """Assert the request parameters for creating a repository."""
        assert mock_request.call_args[1]["endpoint"] == "projects"
        assert mock_request.call_args[1]["data"]["name"] == self.test_repo_name



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_config.const.types import DictConfig
from wexample_filestate.testing.test_abstract_operation import TestAbstractOperation
from wexample_filestate_git.test.mixin.test_git_state_manager_mixin import (
    TestGitFileStateManagerMixin,
)

if TYPE_CHECKING:
    from wexample_config.const.types import DictConfig


class TestGitRemoteAddOperation(TestGitFileStateManagerMixin, TestAbstractOperation):
    def _operation_get_count(self) -> int:
        return 2

    def _operation_test_assert_applied(self) -> None:
        self._assert_dir_exists(self._get_git_dir_path("test_git_dir"))

    def _operation_test_assert_initial(self) -> None:
        self._assert_dir_exists(self._get_git_dir_path("test_git_dir"), positive=False)

    def _operation_test_setup_configuration(self) -> DictConfig | None:
        from wexample_filestate.const.disk import DiskItemType

        self._remove_test_git_dir()

        return {
            "children": [
                {
                    "name": "test_git_dir",
                    "should_exist": True,
                    "type": DiskItemType.DIRECTORY,
                    "git": {
                        "remote": [
                            {
                                "name": "github",
                                "url": lambda target: f"test-remote-with-{target.get_path().name}",
                            },
                        ]
                    },
                }
            ]
        }



from __future__ import annotations


class GitRemoteTest:
    """Base class for Git remote service tests."""

    # Service specific attributes to be overridden by child classes
    ENV_TOKEN_NAME = None
    SERVICE_CLASS = None
    TEST_TOKEN = "test_token"

    @property
    def test_description(self) -> str:
        """Return the test repository description."""
        return "Test description"

    @property
    def test_namespace(self) -> str:
        """Return the test namespace."""
        return "test-namespace"

    @property
    def test_repo_name(self) -> str:
        """Return the test repository name."""
        return "test-repo"

    def _assert_check_repository_exists_request(self, mock_request) -> None:
        """Assert the request parameters for checking repository existence.
        To be implemented by child classes."""
        raise NotImplementedError()

    def _assert_create_repository_request(self, mock_request) -> None:
        """Assert the request parameters for creating a repository.
        To be implemented by child classes."""
        raise NotImplementedError()



from __future__ import annotations

from typing import TYPE_CHECKING

from wexample_filestate.operations_provider.abstract_operations_provider import (
    AbstractOperationsProvider,
)

if TYPE_CHECKING:
    from wexample_filestate.operation.abstract_operation import AbstractOperation


class GitOperationsProvider(AbstractOperationsProvider):
    @staticmethod
    def get_operations() -> list[type[AbstractOperation]]:
        from wexample_filestate_git.operation.git_create_branch_operation import (
            GitCreateBranchOperation,
        )
        from wexample_filestate_git.operation.git_init_operation import GitInitOperation
        from wexample_filestate_git.operation.git_remote_add_operation import (
            GitRemoteAddOperation,
        )
        from wexample_filestate_git.operation.git_remote_create_operation import (
            GitRemoteCreateOperation,
        )

        return [
            # filestate: python-iterable-sort
            GitCreateBranchOperation,
            GitInitOperation,
            GitRemoteAddOperation,
            GitRemoteCreateOperation,
        ]
